#include "utils.h"

/**
 * @brief definition of function Any2Int
 *
 * @param value any type value
 *
 * @return ZYM::int2048
 *
 * @throw FatalError
 *
 * @details convert any type value to ZYM::int2048
 */
inline ZYM::int2048 Any2Int(const std::any &value) {
  if (auto ptr = std::any_cast<ZYM::int2048>(&value))
    return *ptr;
  else if (auto ptr = std::any_cast<double>(&value)) {
    std::stringstream buf;
    buf << std::fixed << std::setprecision(0) << std::floor(*ptr);
    std::string tmp;
    buf >> tmp;
    return std::move(ZYM::int2048(tmp));
  } else if (auto ptr = std::any_cast<bool>(&value))
    return std::move(ZYM::int2048((long long)(*ptr)));
  else if (auto ptr = std::any_cast<std::string>(&value)) {
    std::string str = *ptr;
    size_t dot_position = str.find('.');
    if (dot_position != std::string::npos) {
      std::string integer_part = str.substr(0, dot_position);
      str = integer_part;
    }
    return std::move(ZYM::int2048(str));
  } else if (auto ptr = std::any_cast<NoneType>(&value))
    return std::move(ZYM::int2048(0));
  else {
    std::cerr << value.type().name() << std::endl;
    throw FatalError("Any2Int2048: unknown type", 29);
  }
}

/**
 * @brief definition of function Any2Float
 *
 * @param value any type value
 *
 * @return double
 *
 * @throw FatalError
 *
 * @details convert any type value to double
 */
double Any2Float(const std::any &value) {
  if (auto ptr = std::any_cast<double>(&value))
    return *ptr;
  else if (auto ptr = std::any_cast<ZYM::int2048>(&value)) {
    double res = 0;
    std::stringstream buf;
    buf << *ptr;
    std::string tmp;
    buf >> tmp;
    int p = 0;
    if (tmp[0] == '-') p++;
    for (; p < tmp.size(); p++) res = res * 10 + tmp[p] - '0';
    if (tmp[0] == '-') res = -res;
    return res;
  } else if (auto ptr = std::any_cast<bool>(&value))
    return (double)(*ptr);
  else if (auto ptr = std::any_cast<std::string>(&value)) {
    double tmp = std::stod(*ptr);
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else
    throw FatalError("Any2Float: unknown type", 30);
}

/**
 * @brief definition of function Any2String
 *
 * @param value any type value
 *
 * @return std::string
 *
 * @throw FatalError
 *
 * @details convert any type value to std::string
 */
std::string Any2String(const std::any &value) {
  std::stringstream buf;
  std::string res;
  if (auto ptr = std::any_cast<double>(&value)) {
    buf << std::fixed << std::setprecision(6) << *ptr;
  } else if (auto ptr = std::any_cast<bool>(&value)) {
    if (*ptr)
      buf << "True";
    else
      buf << "False";
  } else if (auto ptr = std::any_cast<ZYM::int2048>(&value))
    buf << *ptr;
  else if (auto ptr = std::any_cast<std::string>(&value))
    buf << *ptr;
  else if (auto ptr = std::any_cast<NoneType>(&value))
    buf << "None";
  else
    throw FatalError("Any2String: unknown type", 31);
  std::getline(buf, res);
  return res;
}

/**
 * @brief definition of function Any2Bool
 *
 * @param value any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details convert any type value to bool
 */
bool Any2Bool(const std::any &value) {
  if (auto ptr = std::any_cast<ZYM::int2048>(&value)) {
    return (*ptr) != 0;
  } else if (auto ptr = std::any_cast<double>(&value)) {
    return (*ptr) != 0;
  } else if (auto ptr = std::any_cast<bool>(&value)) {
    return *ptr;
  } else if (auto ptr = std::any_cast<std::string>(&value)) {
    return (*ptr) != "";
  } else if (auto ptr = std::any_cast<NoneType>(&value))
    return false;
  else
    throw FatalError("Any2Bool: unknown type", 32);
}

/**
 * @brief definition of function DeQuate
 *
 * @param val any type value
 *
 * @param Variables VariableContainer
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details replace all RawVarible in val with its value
 */
std::any DeQuate(std::any val, VariableContainer &Variables) {
  if (auto ptr = std::any_cast<std::vector<std::any>>(&val)) {
    for (int i = 0; i < ptr->size(); i++)
      (*ptr)[i] = DeQuate((*ptr)[i], Variables);
    return *ptr;
  } else if (auto ptr = std::any_cast<RawVarible>(&val)) {
    return Variables.ReadVariable(ptr->name);
  } else if (std::any_cast<double>(&val) || std::any_cast<ZYM::int2048>(&val) ||
             std::any_cast<bool>(&val) || std::any_cast<std::string>(&val) ||
             std::any_cast<NoneType>(&val))
    return val;
  else
    return NoneType();
  // throw FatalError("DeQuate: unknown type");
}

/**
 * @brief definition of function ConverToSameArithType
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @param allow_string_operation bool
 *
 * @return int
 *
 * @throw FatalError
 *
 * @details convert a and b to the same type
 */
int ConverToSameArithType(std::any &a, std::any &b,
                          bool allow_string_operation) {
  if (std::any_cast<NoneType>(&a) != nullptr &&
      std::any_cast<std::string>(&b) == nullptr)
    a = (bool)false;
  if (std::any_cast<NoneType>(&b) != nullptr &&
      std::any_cast<std::string>(&a) == nullptr)
    b = (bool)false;
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((!allow_string_operation) && (ptr_a_string || ptr_b_string))
    throw InterpretException(
        "ConverToSameArithType: string operation not allowed in this "
        "situation",
        33);
  if (allow_string_operation && (ptr_a_string || ptr_b_string)) {
    return 1;
  }
  int level_a = (ptr_a_bool != nullptr ? 1 : 0) +
                (ptr_a_int2048 != nullptr ? 2 : 0) +
                (ptr_a_float != nullptr ? 3 : 0);
  int level_b = (ptr_b_bool != nullptr ? 1 : 0) +
                (ptr_b_int2048 != nullptr ? 2 : 0) +
                (ptr_b_float != nullptr ? 3 : 0);
  int res_level = std::max(level_a, level_b);
  switch (res_level) {
    case 1:
      break;
    case 2:
      a = Any2Int(a);
      b = Any2Int(b);
      break;
    case 3:
      a = Any2Float(a);
      b = Any2Float(b);
      break;
    default:
      throw FatalError("ConverToSameArithType: unknown type", 36);
  }
  return 0;
}

/**
 * @brief definition of function Add
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details add a and b
 */
std::any Add(std::any a, std::any b) {
  int status = ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_string != nullptr ? 1 : 0) + (ptr_b_string != nullptr ? 1 : 0) ==
      2)
    return (*ptr_a_string) + (*ptr_b_string);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return ZYM::int2048(int(*ptr_a_bool) + int(*ptr_b_bool));
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) + (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr)) {
    double tmp = (*ptr_a_float) + (*ptr_b_float);
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else if ((ptr_a_string != nullptr) || (ptr_b_string != nullptr))
    return Any2String(a) + Any2String(b);
  else
    throw FatalError("Add: Type Error", 38);
}

/**
 * @brief definition of function Sub
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details sub a and b
 */
std::any Sub(std::any a, std::any b) {
  if (std::any_cast<std::string>(&a) != nullptr) a = ZYM::int2048(0);
  if (std::any_cast<std::string>(&b) != nullptr) b = ZYM::int2048(0);
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  ConverToSameArithType(a, b);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return ZYM::int2048(int(*ptr_a_bool) - int(*ptr_b_bool));
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) - (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr)) {
    double tmp = (*ptr_a_float) - (*ptr_b_float);
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else
    throw FatalError("Sub: Type Error", 39);
  return a;
}

/**
 * @brief definition of function Mul
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details mul a and b
 */
std::any Mul(std::any a, std::any b) {
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  int status = ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_string != nullptr ? 1 : 0) + (ptr_b_string != nullptr ? 1 : 0) ==
      2) {
    return std::string("");
  } else if ((ptr_a_string != nullptr ? 1 : 0) +
                 (ptr_b_string != nullptr ? 1 : 0) ==
             1) {
    if (ptr_a_string == nullptr) {
      std::swap(a, b);
      ptr_a_bool = std::any_cast<bool>(&a);
      ptr_b_bool = std::any_cast<bool>(&b);
      ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
      ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
      ptr_a_float = std::any_cast<double>(&a);
      ptr_b_float = std::any_cast<double>(&b);
      ptr_a_string = std::any_cast<std::string>(&a);
      ptr_b_string = std::any_cast<std::string>(&b);
    }
    if (Less(b, ZYM::int2048(0))) {
      b = Neg(b);
      ptr_b_bool = std::any_cast<bool>(&b);
      ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
      ptr_b_float = std::any_cast<double>(&b);
      ptr_b_string = std::any_cast<std::string>(&b);
    }
    if (ptr_b_float != nullptr) {
      b = Any2Int(*ptr_b_float);
      ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
      goto st;
    }
    if (ptr_b_bool != nullptr) {
      if (*ptr_b_bool) {
        return *ptr_a_string;
      } else {
        return std::string("");
      }
    }
    if (std::any_cast<NoneType>(&a) != nullptr ||
        std::any_cast<NoneType>(&b) != nullptr)
      return std::string("");
    if (ptr_b_int2048 != nullptr) {
    st:;
      std::string res;
      for (int i = 0; i < (*ptr_b_int2048); i++) res += *ptr_a_string;
      return res;
    }
  }
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return ZYM::int2048(int(*ptr_a_bool) * int(*ptr_b_bool));
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) * (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr)) {
    double tmp = (*ptr_a_float) * (*ptr_b_float);
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else
    throw FatalError("Mul: Type Error", 42);
}

/**
 * @brief definition of function Div
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details div a and b
 */
std::any Div(std::any a, std::any b) {
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_string != nullptr) || (ptr_b_string != nullptr))
    throw InterpretException(
        "Div: string operation not allowed in this situation", 43);
  double t_a = Any2Float(a);
  double t_b = Any2Float(b);
  t_a = t_a / t_b;
  if (t_a > -1e-8 && t_a < 0) t_a = 0;
  return t_a;
}

/**
 * @brief definition of function Divv
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details divv a and b
 */
std::any Divv(std::any a, std::any b) {
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  if (std::any_cast<std::string>(&a) != nullptr) a = ZYM::int2048(0);
  if (std::any_cast<std::string>(&b) != nullptr) b = ZYM::int2048(0);
  ConverToSameArithType(a, b);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr)) {
    double tmp = std::floor((*ptr_a_float) / (*ptr_b_float));
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr)) {
      return (*ptr_a_int2048) / (*ptr_b_int2048);
  } else if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr)) {
      return ZYM::int2048(int(*ptr_a_bool) / int(*ptr_b_bool));
  } else
    throw FatalError("Divv: Type Error", 48);
}

/**
 * @brief definition of function Mod
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details mod a and b
 */
std::any Mod(std::any a, std::any b) {
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  if (std::any_cast<std::string>(&a) != nullptr) a = ZYM::int2048(0);
  ConverToSameArithType(a, b);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr)) {
    double tmp = (*ptr_a_float) -
                 (*ptr_b_float) * std::floor((*ptr_a_float) / (*ptr_b_float));
    if (tmp > -1e-8 && tmp < 0) tmp = 0;
    return tmp;
  } else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) % (*ptr_b_int2048);
  else if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return ZYM::int2048(int(*ptr_a_bool) % int(*ptr_b_bool));
  else
    throw FatalError("Mod: Type Error", 52);
}

std::any &SelfAdd(std::any &a, std::any b) {
  a = Add(a, b);
  return a;
}

std::any &SelfSub(std::any &a, std::any b) {
  a = Sub(a, b);
  return a;
}

std::any &SelfMul(std::any &a, std::any b) {
  a = Mul(a, b);
  return a;
}

std::any &SelfDiv(std::any &a, std::any b) {
  a = Div(a, b);
  return a;
}

std::any &SelfDivv(std::any &a, std::any b) {
  a = Divv(a, b);
  return a;
}

std::any &SelfMod(std::any &a, std::any b) {
  a = Mod(a, b);
  return a;
}

/**
 * @brief definition of function Neg
 *
 * @param a any type value
 *
 * @return std::any
 *
 * @throw FatalError
 *
 * @details get -a
 */
std::any Neg(std::any a) {
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  double *ptr_a_float = std::any_cast<double>(&a);
  if (ptr_a_bool != nullptr)
    return ZYM::int2048(-(*ptr_a_bool));
  else if (ptr_a_int2048 != nullptr)
    return -(*ptr_a_int2048);
  else if (ptr_a_float != nullptr)
    return -(*ptr_a_float);
  else
    throw FatalError("Neg: Type Error", 53);
}

/**
 * @brief definition of function Greater
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a>b
 */
bool Greater(std::any a, std::any b) {
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) > (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) > (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) > (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) > (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr))) {
    if (std::any_cast<NoneType>(&b) && *ptr_a_string != "") return true;
    if (ptr_b_bool != nullptr) {
      a = Any2Bool(a);
      return int(std::any_cast<bool>(a)) > int(*ptr_b_bool);
    }
    if (ptr_b_int2048 != nullptr) {
      a = ZYM::int2048(Any2Bool(a));
      return std::any_cast<ZYM::int2048>(a) > *ptr_b_int2048;
    }
    return false;
  } else
    throw FatalError("Greater: Type Error", 54);
}

/**
 * @brief definition of function Less
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a<b
 */
bool Less(std::any a, std::any b) {
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) < (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) < (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) < (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) < (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr))) {
    if (std::any_cast<NoneType>(&a) && *ptr_b_string != "") return true;
    if (ptr_a_bool != nullptr) {
      b = Any2Bool(b);
      return int(*ptr_a_bool) < int(std::any_cast<bool>(b));
    }
    if (ptr_a_int2048 != nullptr) {
      b = ZYM::int2048(Any2Bool(b));
      return *ptr_a_int2048 < std::any_cast<ZYM::int2048>(b);
    }
    return false;
  } else
    throw FatalError("Less: Type Error", 55);
}

/**
 * @brief definition of function Equal
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a==b
 */
bool Equal(std::any a, std::any b) {
  if ((std::any_cast<NoneType>(&a) != nullptr ? 1 : 0) +
          (std::any_cast<NoneType>(&b) != nullptr ? 1 : 0) ==
      1)
    return false;
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) == (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) == (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) == (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) == (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr)))
    return false;
  else
    throw FatalError("Equal: Type Error", 56);
}

/**
 * @brief definition of function NotEqual
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a!=b
 */
bool NotEqual(std::any a, std::any b) {
  if ((std::any_cast<NoneType>(&a) != nullptr ? 1 : 0) +
          (std::any_cast<NoneType>(&b) != nullptr ? 1 : 0) ==
      1)
    return true;
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) != (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) != (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) != (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) != (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr)))
    return true;
  else
    throw FatalError("NotEqual: Type Error", 57);
}

/**
 * @brief definition of function GreaterEqual
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a>=b
 */
bool GreaterEqual(std::any a, std::any b) {
  int cnt = (std::any_cast<NoneType>(&a) != nullptr ? 1 : 0) +
            (std::any_cast<NoneType>(&b) != nullptr ? 1 : 0);
  if (cnt == 2) return true;
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) >= (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) >= (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) >= (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) >= (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr))) {
    if (std::any_cast<NoneType>(&b)) return true;
    if (ptr_b_bool != nullptr) {
      a = Any2Bool(a);
      return int(std::any_cast<bool>(a)) >= int(*ptr_b_bool);
    }
    if (ptr_b_int2048 != nullptr) {
      a = ZYM::int2048(Any2Bool(a));
      return std::any_cast<ZYM::int2048>(a) >= *ptr_b_int2048;
    }
    return false;
  } else
    throw FatalError("GreaterEqual: Type Error", 58);
}

/**
 * @brief definition of function LessEqual
 *
 * @param a any type value
 *
 * @param b any type value
 *
 * @return bool
 *
 * @throw FatalError
 *
 * @details judge if a<=b
 */
bool LessEqual(std::any a, std::any b) {
  int cnt = (std::any_cast<NoneType>(&a) != nullptr ? 1 : 0) +
            (std::any_cast<NoneType>(&b) != nullptr ? 1 : 0);
  if (cnt == 2) return true;
  if (std::any_cast<NoneType>(&a)) a = (bool)false;
  if (std::any_cast<NoneType>(&b)) b = (bool)false;
  ConverToSameArithType(a, b, true);
  bool *ptr_a_bool = std::any_cast<bool>(&a);
  bool *ptr_b_bool = std::any_cast<bool>(&b);
  ZYM::int2048 *ptr_a_int2048 = std::any_cast<ZYM::int2048>(&a);
  ZYM::int2048 *ptr_b_int2048 = std::any_cast<ZYM::int2048>(&b);
  double *ptr_a_float = std::any_cast<double>(&a);
  double *ptr_b_float = std::any_cast<double>(&b);
  std::string *ptr_a_string = std::any_cast<std::string>(&a);
  std::string *ptr_b_string = std::any_cast<std::string>(&b);
  if ((ptr_a_bool != nullptr) && (ptr_b_bool != nullptr))
    return (*ptr_a_bool) <= (*ptr_b_bool);
  else if ((ptr_a_int2048 != nullptr) && (ptr_b_int2048 != nullptr))
    return (*ptr_a_int2048) <= (*ptr_b_int2048);
  else if ((ptr_a_float != nullptr) && (ptr_b_float != nullptr))
    return (*ptr_a_float) <= (*ptr_b_float);
  else if ((ptr_a_string != nullptr) && (ptr_b_string != nullptr))
    return (*ptr_a_string) <= (*ptr_b_string);
  else if (((ptr_a_string != nullptr) || (ptr_b_string != nullptr))) {
    if (std::any_cast<NoneType>(&a)) return true;
    if (ptr_a_bool != nullptr) {
      b = Any2Bool(b);
      return int(*ptr_a_bool) <= int(std::any_cast<bool>(b));
    }
    if (ptr_a_int2048 != nullptr) {
      b = ZYM::int2048(Any2Bool(b));
      return *ptr_a_int2048 <= std::any_cast<ZYM::int2048>(b);
    }
    return false;
  } else
    throw FatalError("LessEqual: Type Error", 59);
}